"""
damp11113-library - A Utils library and Easy to use. For more info visit https://github.com/damp11113/damp11113-library/wiki
Copyright (C) 2021-present damp11113 (MIT)

Visit https://github.com/damp11113/damp11113-library

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
from iso639 import languages
from typing import Any
import numpy as np

def timestamp2date(timestamp, display='%Y-%m-%d %H:%M:%S'):
    return datetime.fromtimestamp(timestamp).strftime(display)

def str2bin(s):
    return ''.join(format(ord(x), '08b') for x in s)

def str2binnparray(s):
    return np.array([int(bit) for bit in list(''.join(format(ord(x), '08b') for x in s))])

def numpyarray2str(arr):
    # Assuming arr is a NumPy array of binary values
    binary_str = ''.join([str(bit) for bit in arr])
    str_data = ''
    for i in range(0, len(binary_str), 8):
        byte = binary_str[i:i+8]
        str_data += chr(int(byte, 2))
    return str_data

def str2bin2(s):
    binary_strings = [format(num, '08b') for num in s]
    return ''.join(binary_strings)

def bin2str(b):
    return ''.join(chr(int(b[i:i+8], 2)) for i in range(0, len(b), 8))

def bytes2bin(data):
    binary_string = ""
    for byte in data:
        binary_string += format(byte, '08b')  # Convert each byte to its binary representation with 8 bits
    return binary_string

def bin2bytes(binary_string):
    bytes_data = bytearray()
    for i in range(0, len(binary_string), 8):
        byte = binary_string[i:i+8]
        bytes_data.append(int(byte, 2))
    return bytes(bytes_data)

def bin2bool(bin):
    bin = list(bin)
    booll = []
    for i in bin:
        i = int(i)
        if i == 0:
            booll.append(False)
        elif i == 1:
            booll.append(True)
    return booll

def rgb2hex(rgb):
    return '%02x%02x%02x' % rgb

def hex2rgb(value):
    value = value.lstrip('#')
    lv = len(value)
    return tuple(int(value[i:i+lv//3], 16) for i in range(0, lv, lv//3))

def text2morse(text, morselang=None):
    if morselang is None:
        morselang = {
            'A': '.-', 'B': '-...', 'C': '-.-.', 'D': '-..', 'E': '.', 'F': '..-.',
            'G': '--.', 'H': '....', 'I': '..', 'J': '.---', 'K': '-.-', 'L': '.-..',
            'M': '--', 'N': '-.', 'O': '---', 'P': '.--.', 'Q': '--.-', 'R': '.-.',
            'S': '...', 'T': '-', 'U': '..-', 'V': '...-', 'W': '.--', 'X': '-..-',
            'Y': '-.--', 'Z': '--..', ' ': ' ', '0': '-----', '1': '.----', '2': '..---',
            '3': '...--', '4': '....-', '5': '.....', '6': '-....', '7': '--...', '8': '---..',
            '9': '----.', '&': '.-...', "'": '.----.', '@': '.--.-.', ')': '-.--.-', '(': '-.--.',
            ':': '---...', ',': '--..--', '=': '-...-', '!': '-.-.--', '.': '.-.-.-', '-': '-....-',
            '+': '.-.-.', '"': '.-..-.', '?': '..--..', '/': '-..-.'
        }

    morse = ''

    for char in text:
        morse += morselang[char.upper()] + ' '
    return morse

def morse2text(morse, morselang=None):
    if morselang is None:
        morselang = {
            'A': '.-', 'B': '-...', 'C': '-.-.', 'D': '-..', 'E': '.', 'F': '..-.',
            'G': '--.', 'H': '....', 'I': '..', 'J': '.---', 'K': '-.-', 'L': '.-..',
            'M': '--', 'N': '-.', 'O': '---', 'P': '.--.', 'Q': '--.-', 'R': '.-.',
            'S': '...', 'T': '-', 'U': '..-', 'V': '...-', 'W': '.--', 'X': '-..-',
            'Y': '-.--', 'Z': '--..', ' ': ' ', '0': '-----', '1': '.----', '2': '..---',
            '3': '...--', '4': '....-', '5': '.....', '6': '-....', '7': '--...', '8': '---..',
            '9': '----.', '&': '.-...', "'": '.----.', '@': '.--.-.', ')': '-.--.-', '(': '-.--.',
            ':': '---...', ',': '--..--', '=': '-...-', '!': '-.-.--', '.': '.-.-.-', '-': '-....-',
            '+': '.-.-.', '"': '.-..-.', '?': '..--..', '/': '-..-.'
        }

    code_elements = morse.split(' ')
    text = ''

    for code in code_elements:
        text += morselang[code]
    return text

# thank devxpy for this code (ref. https://gist.github.com/devxpy/063968e0a2ef9b6db0bd6af8079dad2a)

INSTRUMENTS = [
    'Acoustic Grand Piano',
    'Bright Acoustic Piano',
    'Electric Grand Piano',
    'Honky-tonk Piano',
    'Electric Piano 1',
    'Electric Piano 2',
    'Harpsichord',
    'Clavi',
    'Celesta',
    'Glockenspiel',
    'Music Box',
    'Vibraphone',
    'Marimba',
    'Xylophone',
    'Tubular Bells',
    'Dulcimer',
    'Drawbar Organ',
    'Percussive Organ',
    'Rock Organ',
    'Church Organ',
    'Reed Organ',
    'Accordion',
    'Harmonica',
    'Tango Accordion',
    'Acoustic Guitar (nylon)',
    'Acoustic Guitar (steel)',
    'Electric Guitar (jazz)',
    'Electric Guitar (clean)',
    'Electric Guitar (muted)',
    'Overdriven Guitar',
    'Distortion Guitar',
    'Guitar harmonics',
    'Acoustic Bass',
    'Electric Bass (finger)',
    'Electric Bass (pick)',
    'Fretless Bass',
    'Slap Bass 1',
    'Slap Bass 2',
    'Synth Bass 1',
    'Synth Bass 2',
    'Violin',
    'Viola',
    'Cello',
    'Contrabass',
    'Tremolo Strings',
    'Pizzicato Strings',
    'Orchestral Harp',
    'Timpani',
    'String Ensemble 1',
    'String Ensemble 2',
    'SynthStrings 1',
    'SynthStrings 2',
    'Choir Aahs',
    'Voice Oohs',
    'Synth Voice',
    'Orchestra Hit',
    'Trumpet',
    'Trombone',
    'Tuba',
    'Muted Trumpet',
    'French Horn',
    'Brass Section',
    'SynthBrass 1',
    'SynthBrass 2',
    'Soprano Sax',
    'Alto Sax',
    'Tenor Sax',
    'Baritone Sax',
    'Oboe',
    'English Horn',
    'Bassoon',
    'Clarinet',
    'Piccolo',
    'Flute',
    'Recorder',
    'Pan Flute',
    'Blown Bottle',
    'Shakuhachi',
    'Whistle',
    'Ocarina',
    'Lead 1 (square)',
    'Lead 2 (sawtooth)',
    'Lead 3 (calliope)',
    'Lead 4 (chiff)',
    'Lead 5 (charang)',
    'Lead 6 (voice)',
    'Lead 7 (fifths)',
    'Lead 8 (bass + lead)',
    'Pad 1 (new age)',
    'Pad 2 (warm)',
    'Pad 3 (polysynth)',
    'Pad 4 (choir)',
    'Pad 5 (bowed)',
    'Pad 6 (metallic)',
    'Pad 7 (halo)',
    'Pad 8 (sweep)',
    'FX 1 (rain)',
    'FX 2 (soundtrack)',
    'FX 3 (crystal)',
    'FX 4 (atmosphere)',
    'FX 5 (brightness)',
    'FX 6 (goblins)',
    'FX 7 (echoes)',
    'FX 8 (sci-fi)',
    'Sitar',
    'Banjo',
    'Shamisen',
    'Koto',
    'Kalimba',
    'Bag pipe',
    'Fiddle',
    'Shanai',
    'Tinkle Bell',
    'Agogo',
    'Steel Drums',
    'Woodblock',
    'Taiko Drum',
    'Melodic Tom',
    'Synth Drum',
    'Reverse Cymbal',
    'Guitar Fret Noise',
    'Breath Noise',
    'Seashore',
    'Bird Tweet',
    'Telephone Ring',
    'Helicopter',
    'Applause',
    'Gunshot'
]
NOTES = ['C', 'C#', 'D', 'D#', 'E', 'F', 'F#', 'G', 'G#', 'A', 'A#', 'B']
OCTAVES = list(range(11))
NOTES_IN_OCTAVE = len(NOTES)

errors = {
    'program': 'Bad input, please refer this spec-\n'
               'https://www.recordingblogs.com/wiki/midi-program-change-message',
    'notes': 'Bad input, please refer this spec-\n'
             'https://newt.phys.unsw.edu.au/jw/notes.html'
}

def instrument2program(instrument: str) -> int:
    assert instrument in INSTRUMENTS, errors['program']
    return INSTRUMENTS.index(instrument) + 1

def program2instrument(program: int) ->  str:
    assert 1 <= program <= 128, errors['program']
    return INSTRUMENTS[program - 1]

def number2note(number: int) -> list[str | int | Any]:
    octave = number // NOTES_IN_OCTAVE
    assert octave in OCTAVES, errors['notes']
    assert 0 <= number <= 127, errors['notes']
    note = NOTES[number % NOTES_IN_OCTAVE]

    return [note, octave]

def note2number(note: str, octave: int) -> int:
    assert note in NOTES, errors['notes']
    assert octave in OCTAVES, errors['notes']

    note = NOTES.index(note)
    note += (NOTES_IN_OCTAVE * octave)

    assert 0 <= note <= 127, errors['notes']

    return note

#------------------------------------------------------------------------

def number2freq(midi_note: int):
    return 2 ** ((midi_note - 69) / 12) * 440

def langpart12fu(lang):
    return languages.get(part1=lang).name

def SI4713RDSPSMISC(stereo=False, artificialhead=False, compressed=False, dynamicpty=False, tp=False, pty=0, forceb=False, ta=False, ms=False):
    # Construct the 16-bit value
    value = (int(stereo) << 15) | (int(artificialhead) << 14) | (int(compressed) << 13) | (int(dynamicpty) << 12) | (int(forceb) << 11) | \
            (int(tp) << 10) | (pty << 5) | (int(ta) << 4) | (int(ms) << 3)

    # Convert the value to hex
    hex_code = hex(value)[2:].zfill(4).upper()

    return hex_code

def number2roman(number):
    if not isinstance(number, int) or not 0 < number < 1000000000000:
        raise ValueError("Input must be an int only and between 1-999999999999 < 1000000000000")

    roman_numerals = [
        (1000000000000, "M̅"), (900000000000, "C̅M̅"), (500000000000, "D̅"), (400000000000, "C̅D̅"),
        (100000000000, "C̅"), (90000000000, "X̅C̅"), (50000000000, "L̅"), (40000000000, "X̅L̅"),
        (10000000000, "X̅"), (9000000000, "M̅X̅"), (5000000000, "V̅"), (4000000000, "M̅V̅"),
        (1000000000, "M̅"), (900000000, "CM"), (500000000, "D"), (400000000, "CD"),
        (100000000, "C"), (90000000, "XC"), (50000000, "L"), (40000000, "XL"),
        (10000000, "X"), (9000000, "IX"), (5000000, "V"), (4000000, "IV"),
        (1000000, "M"), (900000, "CM"), (500000, "D"), (400000, "CD"),
        (100000, "C"), (90000, "XC"), (50000, "L"), (40000, "XL"),
        (10000, "X"), (9000, "IX"), (5000, "V"), (4000, "IV"),
        (1000, "M"), (900, "CM"), (500, "D"), (400, "CD"),
        (100, "C"), (90, "XC"), (50, "L"), (40, "XL"),
        (10, "X"), (9, "IX"), (5, "V"), (4, "IV"),
        (1, "I")
    ]

    roman_numeral = ""
    for value, symbol in roman_numerals:
        while number >= value:
            roman_numeral += symbol
            number -= value

    return roman_numeral

def hammingEncode(data):
    # Check if the input data is a binary string
    if not all(bit in '01' for bit in data):
        raise ValueError("Input data must be a binary string")

    # Determine the number of parity bits required (m)
    m = 0
    while 2 ** m < len(data) + m + 1:
        m += 1

    # Calculate the number of total bits in the codeword
    n = len(data) + m

    # Initialize the encoded codeword with None for parity bit positions
    codeword = [None] * n

    # Copy data bits to their respective positions in the codeword
    data_index = 0
    for i in range(n):
        if i + 1 not in [2 ** j for j in range(m)]:
            codeword[i] = data[data_index]
            data_index += 1

    # Calculate parity bits using XOR
    for i in range(m):
        parity_bit_index = 2 ** i - 1
        xor_result = 0
        for j in range(parity_bit_index, n, 2 ** (i + 1)):
            if codeword[j] is not None:
                xor_result ^= int(codeword[j])
        codeword[parity_bit_index] = str(xor_result)

    return ''.join(codeword)

def hammingDecode(codeword):
    try:
        # Check if the input codeword is a binary string
        if not all(bit in '01' or bit == 'P' for bit in codeword):
            raise ValueError("Input codeword must be a binary string with 'P' placeholders")

        # Determine the number of parity bits (m)
        m = 0
        while 2 ** m < len(codeword) + 1:
            m += 1

        # Calculate syndrome bits
        syndrome = ''
        for i in range(m):
            parity_bit_index = 2 ** i - 1
            xor_result = 0
            for j in range(parity_bit_index, len(codeword), 2 ** (i + 1)):
                if codeword[j] != 'P':
                    xor_result ^= int(codeword[j])
            syndrome = str(xor_result) + syndrome

        # Convert syndrome to error index
        error_index = int(syndrome, 2)

        if error_index == 0:
            # No error
            data_bits = []
            for i in range(len(codeword)):
                if i + 1 not in [2 ** j for j in range(m)]:
                    data_bits.append(codeword[i])
            data = ''.join(data_bits)
        else:
            # Correct the error
            corrected_bit = '1' if codeword[error_index - 1] == '0' else '0'
            corrected_codeword = codeword[:error_index - 1] + corrected_bit + codeword[error_index:]
            data = hammingDecode(corrected_codeword)

        return data
    except:
        return "0"

def nparray2bin(nparray):
    binary_string = ''.join(str(bit) for bit in nparray)
    return binary_string

# Convert binary string to an array of binary values
def bin2nparray(binary_string):
    binary_array = np.array([int(bit) for bit in binary_string], dtype=np.uint8)
    return binary_array

def sample2bar(sample, sensitivity=2, bar='|'):
    peak = np.average(np.abs(sample)) * sensitivity
    bars = bar * int(50 * peak / 2 ** 16)
    return bars

def calculate_shifted_frequency(center_freq, shift):
    censhift = shift / 2
    tone1 = center_freq - censhift
    tone2 = center_freq + censhift
    return tone1, tone2

# Define dictionaries mapping letters and numbers to their phonetic representations
phonetic_alphabet = {
    'A': 'Alpha', 'B': 'Bravo', 'C': 'Charlie', 'D': 'Delta', 'E': 'Echo',
    'F': 'Foxtrot', 'G': 'Golf', 'H': 'Hotel', 'I': 'India', 'J': 'Juliett',
    'K': 'Kilo', 'L': 'Lima', 'M': 'Mike', 'N': 'November', 'O': 'Oscar',
    'P': 'Papa', 'Q': 'Quebec', 'R': 'Romeo', 'S': 'Sierra', 'T': 'Tango',
    'U': 'Uniform', 'V': 'Victor', 'W': 'Whiskey', 'X': 'X-ray', 'Y': 'Yankee', 'Z': 'Zulu',
    '0': 'Zero', '1': 'One', '2': 'Two', '3': 'Three', '4': 'Four',
    '5': 'Five', '6': 'Six', '7': 'Seven', '8': 'Eight', '9': 'Nine'
}

# Function to translate a word or number into phonetic representation
def str2phonetic(input_string):
    # Convert the input string to uppercase for consistency
    input_string = input_string.upper()
    # Translate each character to phonetic representation and join them
    translated_string = ' '.join(phonetic_alphabet.get(char, char) for char in input_string)
    return translated_string

# Function to translate phonetic representation to a word or number
def phonetic2str(phonetic):
    # Split phonetic representation into words and translate each word back to character or number
    words = phonetic.split()
    translated_string = ''.join([key for word in words for key, value in phonetic_alphabet.items() if value == word])
    return translated_string


def str2bool(s):
    true_values = {'true', 'yes', '1', 'on', 'y', 't'}
    false_values = {'false', 'no', '0', 'off', 'n', 'f'}

    s = s.strip().lower()

    if s in true_values:
        return True
    elif s in false_values:
        return False
    else:
        raise ValueError(f"Invalid boolean string: {s}")





"""
damp11113-library - A Utils library and Easy to use. For more info visit https://github.com/damp11113/damp11113-library/wiki
Copyright (C) 2021-present damp11113 (MIT)

Visit https://github.com/damp11113/damp11113-library

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import gc; gc.enable()
import os
import platform

from .info import __version__

try:
    os.environ["damp11113_load_all_module"]
except:
    os.environ["damp11113_load_all_module"] = "YES"

try:
    os.environ["damp11113_check_update"]
except:
    os.environ["damp11113_check_update"] = "YES"

if os.environ["damp11113_load_all_module"] == "YES":
    if platform.system() == "Windows":
        from .pywindows import *

    from .info import *
    from .file import *
    from .randoms import *
    from .processbar import *
    from .convert import *
    from .utils import *
    from .plusmata import *
    from .logic import *
    from .PyserHTTP import *
    from .format import *
    from .imageps import *
    from .minecraft import *
    from .media import *
    from .network import *
    from .DSP import *
    from .codec import *

    from .OPFONMW.dearpygui_animate import *
    from .OPFONMW.StepperLib import *

if os.environ["damp11113_check_update"] == "YES":
    import threading
    import time
    def check_for_update():
        from pygments import console
        import requests
        if os.name == 'nt':
            import ctypes
            kernel32 = ctypes.windll.kernel32
            kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)

        print(console.colorize("yellow", "Checking for damp11113-library update..."))
        try:
            # Attempt to fetch version info with a timeout and retry logic
            retries = 3
            for attempt in range(retries):
                try:
                    response = requests.get("https://cdn.damp11113.xyz/file/text/damp11113libver.txt", timeout=5)
                    if response.status_code == 200:
                        # Compare versions and display results
                        if response.text == __version__:
                            print(console.colorize("green", f"No update available for damp11113-library."))
                            print(console.colorize("green", f"damp11113-library version: {__version__}"))
                        else:
                            print(console.colorize("yellow", "Update available!"))
                            print(console.colorize("green", f"Current version: {__version__}"))
                            print(console.colorize("green", f"New version: {response.text}"))
                        break  # Exit after handling the response
                    else:
                        print(console.colorize("red", f"Error {response.status_code}: Update check failed."))
                        break
                except requests.exceptions.RequestException as e:
                    if attempt < retries - 1:
                        print(console.colorize("yellow", f"Retrying... ({attempt + 1}/{retries})"))
                        time.sleep(2)  # Wait before retrying
                    else:
                        print(console.colorize("red", f"Failed to check for update: {e}"))
        except Exception as e:
            print(console.colorize("red", f"Unexpected error: {e}"))


    # Create and start the thread for the update check
    update_thread = threading.Thread(target=check_for_update)
    update_thread.start()


from setuptools import setup, find_packages
import platform
import os

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

# Common dependencies for all platforms
common_dependencies = [
    "dearpygui",
    "iso-639",
    "numpy",
    "scipy",
    "natsort",
    "psutil",
    "Pillow",
    "opencv-python",
    "libscrc",
    "tqdm",
    "qrcode",
    "python-barcode",
    "pydub",
    "pyzbar",
    "paho-mqtt",
    "requests",
    "pymata-aio",
    "PyQt5",
    "py-cpuinfo",
    "GPUtil",
    "matplotlib",
    "bitarray",
    "scikit-learn",
    "pyaudio",
    "pygments",
]

# Windows-specific dependencies
windows_dependencies = [
    "pywin32",
    "comtypes",
]

# Add platform-specific dependencies
if platform.system() == 'Windows':
    install_requires = common_dependencies + windows_dependencies
else:
    install_requires = common_dependencies

setup(
    name='damp11113',
    version='2025.2.22',
    license='MIT',
    author='damp11113',
    author_email='damp51252@gmail.com',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/damp11113/damp11113-library',
    description="A Utils library and Easy to using.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=install_requires,
)



"""
damp11113-library - A Utils library and Easy to use. For more info visit https://github.com/damp11113/damp11113-library/wiki
Copyright (C) 2021-present damp11113 (MIT)

Visit https://github.com/damp11113/damp11113-library

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import os
import re
import zipfile
import json
import natsort
import psutil
import configparser

#-----------------------------read---------------------------------------

def quickRead(file, decode='utf-8'):
    with open(file, 'r', encoding=decode) as f:
        return f.read()

def quickReadJson(file):
    with open(file, 'r', encoding='utf-8') as f:
        return json.load(f)

def readini(filename, section_name, parameter_name):
    config = configparser.ConfigParser()
    config.read(filename)

    if section_name in config and parameter_name in config[section_name]:
        return config[section_name][parameter_name]
    else:
        print(f"Section '{section_name}' or parameter '{parameter_name}' not found in the INI file.")
        return None

#----------------------------------write------------------------------------

def quickWrite(file, data, encode="utf-8"):
    with open(file, 'w', encoding=encode) as f:
        f.write(data)
        f.close()

def quickWriteJson(file, data):
    with open(f'{file}.json', 'w') as f:
        json.dump(data, f)

#--------------------------------------zip----------------------------------

def unzip(file, to):
    with zipfile.ZipFile(file, 'r') as zip_ref:
        zip_ref.extractall(to)

def comzip(file, to):
    with zipfile.ZipFile(file, 'w') as zip_ref:
        zip_ref.write(to)

#----------------------------------size------------------------------------

def sizefile(file):
    size = 0

    for path, dirs, files in os.walk(file):
        for f in files:
            size += os.path.getsize(os.path.join(path, f))

    return size / 1000000

def sizefolder(folder):
    size = 0

    for path, dirs, files in os.walk(folder):
        for f in files:
            fp = os.path.join(path, f)
            size += os.path.getsize(fp)

    return size / 1000000

def sizefolder2(folder):
    size = psutil.disk_usage(folder).used
    return size / 1000000

def sizefolder3(folder):
    size = 0

    for path, dirs, files in os.walk(folder):
        for f in files:
            fp = os.path.join(path, f)
            size += os.path.getsize(fp)

    return size

#----------------------------------all-------------------------------------

def allfiles(folder, scan_subfolders=False, include_path=False, sort_by=None, sort_reverse=False, valid_extensions=None):
    """
    Scans files in a specified folder (and optionally its subfolders), and returns a list of file names
    with options for sorting and including the full file path.

    Parameters:
    - folder (str): The path of the folder to scan for files.
    - scan_subfolders (bool): If True, recursively scans subfolders. Defaults to False.
    - include_path (bool): If True, includes the full path of each file. Defaults to False.
    - sort_by (str or None): Defines the sorting criterion. Options are:
        - 'name': Sort files alphabetically.
        - 'size': Sort files by size.
        - 'created': Sort files by creation time.
        - 'modified': Sort files by last modified time.
        Defaults to None (no sorting).
    - sort_reverse (bool): If True, sorts in reverse order (descending). Defaults to False.
    - valid_extensions (tuple or None): A tuple of file extensions to include (e.g., ('.mp4', '.jpg')). If None, no filter is applied.

    Returns:
    - list: A list of file names or paths, sorted as per the specified options.

    Example:
    - allfiles("/path/to/folder", scan_subfolders=True, include_path=True, sort_by="name", sort_reverse=True, valid_extensions=('.mp4', '.jpg'))
    """

    # If valid_extensions is None, disable the filter and return all files
    if valid_extensions is None:
        valid_extensions = []

    all_files = []

    # Scan files in subfolders if needed
    if scan_subfolders:
        for root, dirs, files in os.walk(folder):
            for file in files:
                # If valid_extensions is empty, no filtering is applied
                if not valid_extensions or file.endswith(valid_extensions):
                    file_path = os.path.join(root, file) if include_path else file
                    all_files.append(file_path)
    else:
        for file in os.listdir(folder):
            # If valid_extensions is empty, no filtering is applied
            if not valid_extensions or file.endswith(valid_extensions):
                file_path = os.path.join(folder, file) if include_path else file
                all_files.append(file_path)

    # Sorting the files if requested
    if sort_by:
        # Sorting options
        if sort_by == 'name':
            all_files.sort(key=lambda x: [int(text) if text.isdigit() else text.lower() for text in re.split('(\d+)', x)], reverse=sort_reverse)  # Sort alphabetically (case insensitive)
        elif sort_by == 'size':
            all_files.sort(key=lambda x: os.path.getsize(x), reverse=sort_reverse)  # Sort by file size
        elif sort_by == 'created':
            all_files.sort(key=lambda x: os.path.getctime(x), reverse=sort_reverse)  # Sort by creation time
        elif sort_by == 'modified':
            all_files.sort(key=lambda x: os.path.getmtime(x), reverse=sort_reverse)  # Sort by last modified time

        return all_files

    return all_files

#----------------------------------count-------------------------------------

def countline(file, decode='utf-8'):
    with open(file, 'r', encoding=decode) as f:
        line = sum(1 for _ in f)
    return line

#-------------------------------sort_files-------------------------

def sort_files(file_list, reverse=False):
    flist = []
    for file in file_list:
        flist.append(file)
    return natsort.natsorted(flist, reverse=reverse)



"""
damp11113-library - A Utils library and Easy to use. For more info visit https://github.com/damp11113/damp11113-library/wiki
Copyright (C) 2021-present damp11113 (MIT)

Visit https://github.com/damp11113/damp11113-library

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import warnings
import json
import requests
from base64 import b64decode as base64decode
import socket
import struct

class mcstatus_exception(Exception):
    pass

class uuid2name_exception(Exception):
    pass

class server_exeption(Exception):
    pass

class install_exception(Exception):
    pass


#------------------get-uuid2name------------------

class uuid2name:
    def __init__(self) -> None:
        pass

    @staticmethod
    def getmcuuid(player_name):
        try:
            r = requests.get(f"https://api.mojang.com/users/profiles/minecraft/{player_name}")
            if r.text == '':
                raise uuid2name_exception(f"{player_name} not found")
            else:
                return json.loads(r.text)['id']
        except Exception as e:
            raise uuid2name_exception(e)

    @staticmethod
    def getmcname(player_uuid):
        try:
            r = requests.get(f"https://api.mojang.com/user/profiles/{player_uuid}/names")
        except Exception as e:
            raise uuid2name_exception(e)
        try:
            o = json.loads(r.text)[0]['name']
            return o
        except KeyError:
            raise uuid2name_exception(f"player not found")

    @staticmethod
    def getmcnamejson(player_uuid):
        try:
            return requests.get(f"https://api.mojang.com/user/profiles/{player_uuid}/names").text
        except Exception as e:
            raise uuid2name_exception(f"get mc name error: {e}")

    @staticmethod
    def getmcuuidjson(player_name):
        try:
            return requests.get(f"https://api.mojang.com/users/profiles/minecraft/{player_name}").text
        except Exception as e:
            raise uuid2name_exception(f"get mc uuid error: {e}")

#----------------other api------------------

def skin_url(uuid):
    try:
        r = requests.get(f"https://sessionserver.mojang.com/session/minecraft/profile/{uuid}")
        base = json.loads(r.text)['properties'][0]['value']
        # base64 to json
        js = base64decode(base)
        # json to dict
        d = json.loads(js)
        # get skin url
        return d['textures']['SKIN']['url']
    except Exception as e:
        raise uuid2name_exception(f"skin url error: {e}")

def mctimestamp(uuid):
    try:
        r = requests.get(f"https://sessionserver.mojang.com/session/minecraft/profile/{uuid}")
        base = json.loads(r.text)['properties'][0]['value']
        # base64 to json
        js = base64decode(base)
        # json to dict
        d = json.loads(js)
        # get skin url
        return d['timestamp']
    except Exception as e:
        raise uuid2name_exception(f"error: {e}")

#----------------------mcstatus------------------------



class mcstatusv2:
    def __init__(self, ip, port=25565):
        """
        To use visit https://chat.openai.com/share/966ad89a-3785-4d94-a4cc-7cb05c73bab3
        """
        warnings.warn("mcstatusv2 is experimental and not fully implemented", Warning)

        self.ip = ip
        self.port = port

    def _send_receive(self, payload):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(5)

        try:
            sock.connect((self.ip, self.port))
            sock.sendall(payload)
            data = sock.recv(4096)
            return data
        except socket.timeout:
            raise mcstatus_exception("Connection timed out")
        except Exception as e:
            raise mcstatus_exception(f"Error: {e}")
        finally:
            sock.close()

    def _parse_status(self, data):
        length = data[0]
        data = data[1:length+1]
        return json.loads(data.decode('utf-8'))

    def _get_status(self):
        payload = b'\x00\x00'  # Packet for getting status
        payload += bytes([len(self.ip)]) + self.ip.encode('utf-8')  # Append server address
        payload += bytes.fromhex('{:04x}'.format(self.port))  # Append server port
        payload += b'\x01'  # Protocol version

        response = self._send_receive(bytes([len(payload)]) + payload)
        return response

    def status(self):
        try:
            response = self._get_status()
            parsed_status = self._parse_status(response)
            return parsed_status
        except mcstatus_exception as e:
            raise mcstatus_exception(f"Error parsing status: {e}")

#----------------------Rcon-----------------------

class RconV2:
    def __init__(self, host, port, password):
        warnings.warn("RconV2 is experimental and not fully implemented", Warning)
        self.host = host
        self.port = port
        self.password = password
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.connect()

    def connect(self):
        self.sock.connect((self.host, self.port))
        packet = struct.pack('<iii', 0, 0, 2) + self.password.encode('utf-8') + b'\x00\x00'
        self.sock.send(struct.pack('<i', len(packet)) + packet)
        size, id, type = struct.unpack('<iii', self.sock.recv(12))
        response = self.sock.recv(size - 8).decode('utf-8')
        if id == -1:
            raise RuntimeError(f"Failed to authenticate: {response.strip()}")

    def send_command(self, command):
        command_packet = struct.pack('<iii', 1, 2, 2) + command.encode('utf-8') + b'\x00\x00'
        self.sock.send(struct.pack('<i', len(command_packet)) + command_packet)
        size, id, type = struct.unpack('<iii', self.sock.recv(12))
        response = self.sock.recv(size - 8).decode('utf-8')
        return response.strip()

    def close(self):
        self.sock.close()



"""
damp11113-library - A Utils library and Easy to use. For more info visit https://github.com/damp11113/damp11113-library/wiki
Copyright (C) 2021-present damp11113 (MIT)

Visit https://github.com/damp11113/damp11113-library

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from shutil import get_terminal_size
from itertools import cycle
import math
import time
from threading import Thread, Lock, Event
from time import sleep
from .utils import get_size_unit2, center_string, TextFormatter, insert_string

class Steps:
    sending = ['[   ]', '[-  ]', '[-- ]', '[---]', '[ --]', '[  -]']
    requesting = ['[   ]', '[-  ]', '[ - ]', '[  -]']
    waiting = ['[   ]', '[-  ]', '[-- ]', '[ --]', '[  -]', '[   ]', '[  -]', '[ --]', '[-- ]', '[-  ]']
    pinging = ['[   ]', '[-  ]', '[ - ]', '[  -]', '[   ]', '[  -]', '[ - ]', '[-  ]', '[   ]']
    receiving = ['[   ]', '[  -]', '[ --]', '[---]', '[-- ]', '[-  ]']
    connecting = ['[   ]', '[  -]', '[ - ]', '[-  ]']

    expand_contract = ['[    ]', '[=   ]', '[==  ]', '[=== ]', '[====]', '[ ===]', '[  ==]', '[   =]', '[    ]']
    rotating_dots = ['.    ', '..   ', '...  ', '.... ', '.....', ' ....', '  ...', '   ..', '    .', '     ']
    bouncing_ball = ['o     ', ' o    ', '  o   ', '   o  ', '    o ', '     o', '    o ', '   o  ', '  o   ', ' o    ', 'o     ']
    left_right_dots = ['[    ]', '[.   ]', '[..  ]', '[... ]', '[....]', '[ ...]', '[  ..]', '[   .]', '[    ]']
    expanding_square = ['[ ]', '[■]', '[■■]', '[■■■]', '[■■■■]', '[■■■]', '[■■]', '[■]', '[ ]']
    spinner = ['|', '/', '-', '\\', '|', '/', '-', '\\']
    zigzag = ['/   ', ' /  ', '  / ', '   /', '  / ', ' /  ', '/   ', '\\   ', ' \\  ', '  \\ ', '   \\', '  \\ ', ' \\  ', '\\   ']
    arrows = ['←  ', '←← ', '←←←', '←← ', '←  ', '→  ', '→→ ', '→→→', '→→ ', '→  ']
    snake = ['[>    ]', '[=>   ]', '[==>  ]', '[===> ]', '[====>]', '[ ===>]', '[  ==>]', '[   =>]', '[    >]']
    loading_bar = ['[          ]', '[=         ]', '[==        ]', '[===       ]', '[====      ]', '[=====     ]', '[======    ]', '[=======   ]', '[========  ]', '[========= ]', '[==========]']


class indeterminateStatus:
    def __init__(self, desc="Loading...", end="[ ✔ ]", timeout=0.1, fail='[ ❌ ]', steps=None):
        self.desc = desc
        self.end = end
        self.timeout = timeout
        self.faill = fail

        self._thread = Thread(target=self._animate, daemon=True)
        if steps is None:
            self.steps = Steps.sending
        else:
            self.steps = steps
        self.done = False
        self.fail = False

    def start(self):
        self._thread.start()
        return self

    def _animate(self):
        for c in cycle(self.steps):
            if self.done:
                break
            print(f"\r{c} {self.desc}", flush=True, end="")
            sleep(self.timeout)

    def __enter__(self):
        self.start()

    def stop(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(f"\r{self.end}", flush=True)

    def stopfail(self):
        self.done = True
        self.fail = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(f"\r{self.faill}", flush=True)

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()

# This class is developed in 3 October 2023 at 5:00 PM
class LoadingProgress:
    def __init__(self, total=100, totalbuffer=None, length=50, fill='█', fillbufferbar='█', desc="Loading...", status="", enabuinstatus=True, end="[ ✔ ]", timeout=0.1, fail='[ ❌ ]', steps=None, unit="it", barbackground="-", shortnum=False, buffer=False, shortunitsize=1000, currentshortnum=False, show=True, clearline=True, indeterminate=False, barcolor="red", bufferbarcolor="white", barbackgroundcolor="black", color=False):
        """
        Simple loading progress bar python
        @param total: change all total
        @param desc: change description
        @param status: change progress status
        @param end: change success progress
        @param timeout: change speed
        @param fail: change error stop
        @param steps: change steps animation
        @param unit: change unit
        @param buffer: enable buffer progress (experiment)
        @param show: show progress bar
        @param indeterminate: indeterminate mode
        @param barcolor: change bar color
        @param bufferbarcolor: change buffer bar color
        @param barbackgroundcolor: change background color
        @param color: enable colorful
        """
        self.desc = desc
        self.end = end
        self.timeout = timeout
        self.faill = fail
        self.total = total
        self.length = length
        self.fill = fill
        self.enbuinstatus = enabuinstatus
        self.status = status
        self.barbackground = barbackground
        self.unit = unit
        self.shortnum = shortnum
        self.shortunitsize = shortunitsize
        self.currentshortnum = currentshortnum
        self.printed = show
        self.clearline = clearline
        self.indeterminate = indeterminate
        self.barcolor = barcolor
        self.barbackgroundcolor = barbackgroundcolor
        self.enabuffer = buffer
        self.bufferbarcolor = bufferbarcolor
        self.fillbufferbar = fillbufferbar
        self.totalbuffer = totalbuffer
        self.enacolor = color

        self._thread = Thread(target=self._animate, daemon=True)

        if steps is None:
            self.steps = Steps.sending
        else:
            self.steps = steps

        if self.totalbuffer is None:
            self.totalbuffer = self.total

        self.currentpercent = 0
        self.currentbufferpercent = 0
        self.current = 0
        self.currentbuffer = 0
        self.startime = 0
        self.done = False
        self.fail = False
        self.currentprint = ""

    def start(self):
        self._thread.start()
        self.startime = time.perf_counter()
        return self

    def update(self, i=1):
        self.current += i

    def updatebuffer(self, i=1):
        self.currentbuffer += i

    def _animate(self):
        for c in cycle(self.steps):
            if self.done:
                break

            if not self.indeterminate:
                if self.total != 0 or math.trunc(float(self.currentpercent)) > 100:
                    if self.enabuffer:
                        self.currentpercent = ("{0:.1f}").format(100 * (self.current / float(self.total)))

                        filled_length = int(self.length * self.current // self.total)

                        if self.enacolor:
                            bar = TextFormatter.format_text(self.fill * filled_length, self.barcolor)
                        else:
                            bar = self.fill * filled_length

                        self.currentbufferpercent = ("{0:.1f}").format(
                            100 * (self.currentbuffer / float(self.totalbuffer)))

                        if float(self.currentbufferpercent) >= 100.0:
                            self.currentbufferpercent = 100

                        filled_length_buffer = int(self.length * self.currentbuffer // self.totalbuffer)

                        if filled_length_buffer >= self.length:
                            filled_length_buffer = self.length

                        if self.enacolor:
                            bufferbar = TextFormatter.format_text(self.fillbufferbar * filled_length_buffer,
                                                                  self.bufferbarcolor)
                        else:
                            bufferbar = self.fillbufferbar * filled_length_buffer

                        bar = insert_string(bufferbar, bar)

                        if self.enacolor:
                            bar += TextFormatter.format_text(self.barbackground * (self.length - filled_length_buffer),
                                                            self.barbackgroundcolor)
                        else:
                            bar += self.barbackground * (self.length - filled_length_buffer)
                    else:
                        self.currentpercent = ("{0:.1f}").format(100 * (self.current / float(self.total)))
                        filled_length = int(self.length * self.current // self.total)
                        if self.enacolor:
                            bar = TextFormatter.format_text(self.fill * filled_length, self.barcolor)

                            bar += TextFormatter.format_text(self.barbackground * (self.length - filled_length),
                                                             self.barbackgroundcolor)
                        else:
                            bar = self.fill * filled_length
                            if self.enacolor:
                                bar = TextFormatter.format_text(bar, self.barcolor)
                            bar += self.barbackground * (self.length - filled_length)


                    if self.enbuinstatus:
                        elapsed_time = time.perf_counter() - self.startime
                        speed = self.current / elapsed_time if elapsed_time > 0 else 0
                        remaining = self.total - self.current
                        eta_seconds = remaining / speed if speed > 0 else 0
                        elapsed_formatted = time.strftime('%H:%M:%S', time.gmtime(elapsed_time))
                        eta_formatted = time.strftime('%H:%M:%S', time.gmtime(eta_seconds))
                        if self.shortnum:
                            stotal = get_size_unit2(self.total, '', False, self.shortunitsize, False, '')
                            scurrent = get_size_unit2(self.current, '', False, self.shortunitsize, self.currentshortnum, '')
                        else:
                            stotal = self.total
                            scurrent = self.current

                        if math.trunc(float(self.currentpercent)) > 100:
                            elapsed_time = time.perf_counter() - self.startime
                            elapsed_formatted = time.strftime('%H:%M:%S', time.gmtime(elapsed_time))

                            bar = center_string(self.barbackground * self.length, TextFormatter.format_text("Indeterminate", self.barcolor))

                            self.currentprint = f"{c} {self.desc} | --%|{bar}| {scurrent}/{stotal} | {elapsed_formatted} | {get_size_unit2(speed, self.unit, self.shortunitsize)} | {self.status}"

                        else:
                            self.currentprint = f"{c} {self.desc} | {math.trunc(float(self.currentpercent))}%|{bar}| {scurrent}/{stotal} | {elapsed_formatted}<{eta_formatted} | {get_size_unit2(speed, self.unit, self.shortunitsize)} | {self.status}"
                    else:
                        if self.shortnum:
                            stotal = get_size_unit2(self.total, '', False, self.shortunitsize, False, '')
                            scurrent = get_size_unit2(self.current, '', False, self.shortunitsize, self.currentshortnum, '')
                        else:
                            stotal = self.total
                            scurrent = self.current


                        self.currentprint = f"{c} {self.desc} | {math.trunc(float(self.currentpercent))}%|{bar}| {scurrent}/{stotal} | {self.status}"
                else:
                    elapsed_time = time.perf_counter() - self.startime
                    elapsed_formatted = time.strftime('%H:%M:%S', time.gmtime(elapsed_time))

                    bar = center_string(self.barbackground * self.length, TextFormatter.format_text("Indeterminate", self.barcolor))

                    self.currentprint = f"{c} {self.desc} | --%|{bar}| {elapsed_formatted} | {self.status}"
            else:
                elapsed_time = time.perf_counter() - self.startime
                elapsed_formatted = time.strftime('%H:%M:%S', time.gmtime(elapsed_time))

                bar = center_string(self.barbackground * self.length, TextFormatter.format_text("Indeterminate", self.barcolor))

                self.currentprint = f"{c} {self.desc} | --%|{bar}| {elapsed_formatted} | {self.status}"

            if self.printed:
                print(f"\r{self.currentprint}", flush=True, end="")

            sleep(self.timeout)

            if self.printed and self.clearline:
                # This clears the previous printed line
                print("\r" + " " * len(self.currentprint), end="", flush=True)

    def __enter__(self):
        self.start()

    def stop(self):
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(f"\r{self.end}", flush=True)

    def stopfail(self):
        self.done = True
        self.fail = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(f"\r{self.faill}", flush=True)

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()

# ------------------------------------------------------------------

class BarPlus_StackedBar:
    """Represents a segment in a stacked progress bar"""

    def __init__(self, value, max_value, color, label="", layer=0):
        self.value = value
        self.max_value = max_value
        self.color = color
        self.label = label
        self.layer = layer  # Layer depth: 0 = front, higher numbers = back
        self.percentage = (value / max_value * 100) if max_value > 0 else 0


class BarPlus_ProgressBar:
    """Individual progress bar with support for stacked segments"""

    def __init__(self, label, width=40, show_percentage=True, show_values=True, steps_animation=None, ena_steps=True,
                 fill='█', barbackground="░", ena_stack_layer=False, stopped="[ ✔ ]", stopped_fail='[ ❌ ]'):
        self.label = label
        self.width = width
        self.show_percentage = show_percentage
        self.show_values = show_values
        self.stacked_bars = []
        self.info_text = ""
        self.steps_animation = Steps.sending if not steps_animation else steps_animation
        self.current_steps_ani = 0
        self.total_steps_ani = len(self.steps_animation)
        self.ena_steps = ena_steps
        self.fill = fill
        self.barbackground = barbackground
        self.ena_stack_layer = ena_stack_layer
        self.stopped = stopped
        self.stopped_fail = stopped_fail
        self.is_stopped = False
        self.is_failed = False

    def change_animation_steps(self, steps_animation):
        """Change the animation steps used for rendering"""
        self.steps_animation = steps_animation
        self.total_steps_ani = len(steps_animation)
        self.current_steps_ani = 0

    def add_segment(self, value, max_value, color, label="", layer=0):
        """Add a segment to the stacked bar with optional layer support"""
        segment = BarPlus_StackedBar(value, max_value, color, label, layer)
        self.stacked_bars.append(segment)

    def update_segment(self, index, value):
        """Update the value of a specific segment"""
        if 0 <= index < len(self.stacked_bars):
            self.stacked_bars[index].value = value
            self.stacked_bars[index].percentage = (
                    value / self.stacked_bars[index].max_value * 100
            ) if self.stacked_bars[index].max_value > 0 else 0

    def set_info(self, text):
        """Set additional info text to display"""
        self.info_text = text

    def stop(self):
        """Stop the progress bar and show success status"""
        self.is_stopped = True
        self.is_failed = False

    def stopfail(self):
        """Stop the progress bar and show failure status"""
        self.is_stopped = True
        self.is_failed = True

    def render(self):
        """Render the progress bar as a string"""
        # If stopped, return the stopped status instead of the normal bar
        if self.is_stopped:
            status = self.stopped_fail if self.is_failed else self.stopped
            return f"{status}"

        if not self.stacked_bars:
            return f"{self.label}: No data"

        # Calculate total percentage for display
        total_percentage = 0
        total_current = 0
        total_max = 0

        for bar in self.stacked_bars:
            total_current += bar.value
            total_max += bar.max_value

        if total_max > 0:
            total_percentage = (total_current / total_max) * 100

        # Build the visual bar
        if self.ena_stack_layer:
            bar_str = self._render_layered_bar(total_max)
        else:
            bar_str = self._render_traditional_bar(total_percentage, total_max)

        # Build the result string with animation at the front
        result_parts = []

        # Add animation at the very front if enabled
        if self.ena_steps:
            result_parts.append(f"{self.steps_animation[self.current_steps_ani]}")
            self.current_steps_ani = (self.current_steps_ani + 1) % self.total_steps_ani

        # Add label
        result_parts.append(f"{self.label}")

        # Add percentage if requested
        if self.show_percentage:
            result_parts.append(f"{total_percentage:3.0f}%")

        # Add the bar itself
        result_parts.append(f"|{bar_str}|")

        # Add segment information
        if self.stacked_bars:
            segments_info = []
            for bar in self.stacked_bars:
                if bar.label:
                    layer_info = f" [L{bar.layer}]" if self.ena_stack_layer else ""
                    segments_info.append(f"{bar.label}: {bar.value}/{bar.max_value}{layer_info}")

            if segments_info:
                result_parts.append(f"{' | '.join(segments_info)}")

        # Add info text
        if self.info_text:
            result_parts.append(f"{self.info_text}")

        # Join all parts with appropriate separators
        return f"{result_parts[0]} {result_parts[1]} | {result_parts[2]} {result_parts[3]} {' '.join(result_parts[4:])}"

    def _render_layered_bar(self, total_max):
        """Render bar with layer support - segments are rendered back to front"""
        # Create array to hold the final bar characters
        bar_chars = [' '] * self.width

        # Sort segments by layer (highest layer first, so they render in back)
        sorted_segments = sorted(self.stacked_bars, key=lambda x: x.layer, reverse=True)

        for segment in sorted_segments:
            if segment.value <= 0:
                continue

            # Calculate the segment's width based on its own percentage
            # Each segment fills independently based on its value/max_value ratio
            segment_width = int((segment.value / segment.max_value) * self.width) if segment.max_value > 0 else 0

            # For layered rendering, each segment starts from the beginning
            # and fills up to its value, overlaying previous layers
            for i in range(min(segment_width, self.width)):
                colored_char = TextFormatter.format_text_truecolor(self.fill, segment.color)
                bar_chars[i] = colored_char

        # Fill remaining space with background
        for i in range(self.width):
            if bar_chars[i] == ' ':
                bar_chars[i] = self.barbackground

        return ''.join(bar_chars)

    def _render_traditional_bar(self, total_percentage, total_max):
        """Render bar with traditional stacking - segments are placed side by side"""
        filled_chars = int((total_percentage / 100) * self.width)
        empty_chars = self.width - filled_chars

        # Create stacked visual representation
        bar_parts = []
        remaining_filled = filled_chars

        for i, bar in enumerate(self.stacked_bars):
            if remaining_filled <= 0:
                break

            # Calculate how much of this segment should be filled
            segment_percentage = bar.percentage
            segment_filled = int((segment_percentage / 100) * (bar.max_value / total_max * self.width))
            segment_filled = min(segment_filled, remaining_filled)

            if segment_filled > 0:
                # Use foreground color for the block characters
                bar_parts.append(TextFormatter.format_text_truecolor(self.fill * segment_filled, bar.color))
                remaining_filled -= segment_filled

        # Add empty space
        if empty_chars > 0:
            bar_parts.append(self.barbackground * empty_chars)

        return ''.join(bar_parts)


class BarPlus_Display:
    """Manages multiple progress bars with terminal positioning"""

    def __init__(self, title="Progress bar", update_rate=0.1):
        self.title = title
        self.bars = []
        self.static_lines = []
        self.terminal_width = get_terminal_size((80, 20)).columns
        self.terminal_height = get_terminal_size((80, 20)).lines
        self.lock = Lock()
        self.start_row = 1
        self.update_rate = update_rate

        # Threading controls
        self._stop_event = Event()
        self._thread = None
        self._is_running = False

    def add_static_line(self, text):
        """Add a static line that appears at the top"""
        with self.lock:
            self.static_lines.append(text)

    def add_progress_bar(self, bar):
        """Add a progress bar to the display"""
        with self.lock:
            self.bars.append(bar)

    def remove_progress_bar(self, bar):
        """Remove a progress bar from the display"""
        with self.lock:
            if bar in self.bars:
                self.bars.remove(bar)

    def clear_static_lines(self):
        """Clear all static lines"""
        with self.lock:
            self.static_lines.clear()

    def move_cursor_to(self, row, col=0):
        """Move cursor to specific position"""
        return f"\033[{row};{col}H"

    def clear_line(self):
        """Clear current line"""
        return "\033[2K"

    def hide_cursor(self):
        """Hide terminal cursor"""
        return "\033[?25l"

    def show_cursor(self):
        """Show terminal cursor"""
        return "\033[?25h"

    def save_cursor(self):
        """Save cursor position"""
        return "\033[s"

    def restore_cursor(self):
        """Restore cursor position"""
        return "\033[u"

    def render(self):
        """Render the complete display"""
        with self.lock:
            output = []
            current_row = self.start_row

            # Add title
            if self.title:
                output.append(self.move_cursor_to(current_row))
                output.append(self.clear_line())
                output.append(TextFormatter.format_text(self.title, attributes='bold'))
                current_row += 1

                # Empty line after title
                output.append(self.move_cursor_to(current_row))
                output.append(self.clear_line())
                current_row += 1

            # Add static lines
            for line in self.static_lines:
                output.append(self.move_cursor_to(current_row))
                output.append(self.clear_line())
                # Truncate if too long
                if len(line) > self.terminal_width:
                    line = line[:self.terminal_width - 3] + "..."
                output.append(line)
                current_row += 1

            # Empty line after static lines
            if self.static_lines:
                output.append(self.move_cursor_to(current_row))
                output.append(self.clear_line())
                current_row += 1

            # Add progress bars
            for bar in self.bars:
                output.append(self.move_cursor_to(current_row))
                output.append(self.clear_line())
                rendered_bar = bar.render()
                # Truncate if too long
                # if len(rendered_bar) > self.terminal_width:
                #     rendered_bar = rendered_bar[:self.terminal_width - 3] + "..."
                output.append(rendered_bar)
                current_row += 1

            return ''.join(output)

    def start(self):
        """Start the display loop"""
        if self._is_running:
            return False  # Already running

        self._stop_event.clear()
        self._is_running = True
        self._thread = Thread(target=self._update_loop, daemon=True)
        self._thread.start()
        return True

    def stop(self):
        """Stop the display loop"""
        if not self._is_running:
            return False  # Not running

        self._stop_event.set()
        self._is_running = False

        # Wait for thread to finish (with timeout)
        if self._thread and self._thread.is_alive():
            self._thread.join(timeout=1.0)

        return True

    def exit(self):
        """Exit and cleanup the display"""
        self.stop()
        self.cleanup()

    def is_running(self):
        """Check if the display loop is running"""
        return self._is_running

    def _update_loop(self):
        """Update the display periodically"""
        try:
            while not self._stop_event.is_set():
                self.display()

                # Use wait instead of sleep for better responsiveness
                if self._stop_event.wait(timeout=self.update_rate):
                    break  # Stop event was set

        except Exception as e:
            #print(f"Error in display update loop: {e}")
            raise e
        finally:
            self._is_running = False

    def display(self):
        """Display the progress bars"""
        print(self.hide_cursor() + self.render(), end='', flush=True)

    def cleanup(self):
        """Clean up display"""
        print(self.show_cursor(), flush=True)

    def __enter__(self):
        """Context manager entry"""
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Context manager exit"""
        self.exit()

    def __del__(self):
        """Destructor - ensure cleanup"""
        if hasattr(self, '_is_running') and self._is_running:
            self.exit()


"""
damp11113-library - A Utils library and Easy to use. For more info visit https://github.com/damp11113/damp11113-library/wiki
Copyright (C) 2021-present damp11113 (MIT)

Visit https://github.com/damp11113/damp11113-library

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import random
import string
import warnings

def rannum(number1, number2):
    try:
        return random.randint(int(number1), int(number2))
    except ValueError:
        print("Please enter a number1 to number2")

def rannumfloat(number1, number2):
    try:
        return random.uniform(number1, number2)
    except ValueError:
        print("Please enter a number1 to number2")

def ranstr(charset):
    try:
        char_set = string.ascii_uppercase + string.digits
        output = ''.join(random.sample(char_set * int(charset), int(charset)))
        return output
    except ValueError:
        print("Please enter a number charset")

def rancolor():
    """RGB"""
    return (rannum(1, 255), rannum(1, 255), rannum(1, 255))

def rantextuplow(text):
    nct = list(text)
    ct = []
    for i in nct:
        r = rannum(1, 2)
        if r == 1:
            ct.append(str(i).lower())
        elif r == 2:
            ct.append(str(i).upper())
    return ''.join(ct)

def ranstruplow(charset):
    return rantextuplow(ranstr(charset))

def ranlossbin(codeword, error_rate):
    # Convert the error rate to a number of errors to introduce
    num_errors = int(len(codeword) * error_rate)

    # Randomly select bit indices to flip
    error_indices = random.sample(range(len(codeword)), num_errors)

    # Flip the selected bits
    received_codeword = list(codeword)
    for index in error_indices:
        received_codeword[index] = '1' if received_codeword[index] == '0' else '0'

    return ''.join(received_codeword)

def ranlossbytes(data, loss_percentage):
    if not 0 <= loss_percentage <= 100:
        raise ValueError("Loss percentage should be between 0 and 100")

    num_bytes_to_drop = int(len(data) * (loss_percentage / 100))
    drop_indices = set(random.sample(range(len(data)), num_bytes_to_drop))

    # Use list comprehension to filter out the indices to drop
    return bytes(byte for i, byte in enumerate(data) if i not in drop_indices)

def old_ranlossbytes(data, loss_percentage):
    warnings.warn("This function is deprecated. Use ranlossbytes instead.", DeprecationWarning)

    if not 0 <= loss_percentage <= 100:
        raise ValueError("Loss percentage should be between 0 and 100")

    num_bytes_to_drop = int(len(data) * (loss_percentage / 100))
    indices_to_drop = random.sample(range(len(data)), num_bytes_to_drop)

    result = bytearray()
    for i, byte in enumerate(data):
        if i not in indices_to_drop:
            result.append(byte)

    return bytes(result)

def rannumlist(number1, number2, maxrange):
    return [random.randint(number1, number2) for _ in range(maxrange)]

def generate_binary_combinations(width):
    binarys = []
    total_combinations = 2 ** width
    for i in range(total_combinations):
        binarys.append(format(i, '0' + str(width) + 'b'))
    return binarys

def ranChooseWithRate(choices_probabilities):
    """
    Randomly choose an item from the given choices-probabilities dictionary.

    Args:
        choices_probabilities (dict): Dictionary where keys are choices and values are probabilities.

    Returns:
        The randomly chosen item.

    Examples: choices_probabilities = {'A': 30, 'B': 20, 'C': 25, 'D': 25}
    """
    choices = list(choices_probabilities.keys())
    probabilities = list(choices_probabilities.values())

    if sum(probabilities) != 100:
        raise ValueError("Probabilities should add up to 100")

    probabilities = [p / 100 for p in probabilities]
    return random.choices(choices, weights=probabilities, k=1)[0]